'''Defines constant values used by
modules in the logging package.
'''

class LogLevel:
	'''Specifies the severity of a LogEntry.
	Entries below a Logger's minimum severity level
	are still saved to the log file, but are not
	printed to standard output.
	'''
	Debug = 0
	Info = 1
	Warn = 2
	Error = 3
	Count = 4

class LogEntryType:
	'''Specifies the type of data stored by a given
	LogEntry.
	'''
	GeneralText = 0
	Count = 1



import os
from ...logging import Logger
from .directory_tree import DirectoryTree

def _subdirsForDir(directory="."):
	return [os.path.join(directory, subdir) for \
	subdir in os.listdir(directory) if \
	os.path.isdir(os.path.join(directory, subdir))]

kReposKey = "repos"
class DirectoryRunner(object):
	def __init__(self, log):
		self._log = log
		assert isinstance(self._log, Logger)
		self.delegate = None
		self._dirTree = DirectoryTree(kReposKey)

	def _trimmedName(self, nodePath):
		nodeName = os.path.basename(nodePath)
		if not nodeName:
			raise RuntimeError(("Node path {0} trimmed to {1}, "
			"which can't be added to the dictionary").format(nodePath, str(nodeName)))
		return nodeName

	def _recurse(self, treeRootPath, delegate, cursor):
		#Perform "entered node" op
		try:
			delegate.onEnteredNode(treeRootPath, cursor)
		except RuntimeError as e:
			#If the delegate had an exception,
			#skip this node and mark error
			self._log.error("Node delegate failed with error: '{0}', skipping node".format(str(e)))
			return
		#Does this have a .git?
		if os.path.exists(treeRootPath + os.path.sep + ".git"):
			#If true, it's a repo. Perform repo traversal op
			#and quit, since we don't handle repos in repos
			try:
				delegate.onRepoFound(treeRootPath, cursor)
			except RuntimeError as e:
				self._log.error("Couldn't traverse node '{0}': {1}".format(treeRootPath, str(e)))
			return
		else:
			#Does this have subdirectories?
			subDirPaths = _subdirsForDir(treeRootPath)
			for s in subDirPaths:
				#Generate subnodes for each subdirectory
				subNodeName = self._trimmedName(s)
				cursor.addSubNode(subNodeName)
				subNodeCursor = cursor.cursorForSubNode(subNodeName)
				#Perform "found subdirectory" op
				delegate.onSubdirectoryFound(s, subNodeCursor)
				#Recurse on subdirectories
				self._recurse(s, delegate, subNodeCursor)

	def directoryStructure(self):
		return self._dirTree.directoryStructure

	def run(self, rootDir, delegate):
		assert delegate
		self.delegate = delegate
		self._dirTree.resetState()
		self.delegate.onStartTraverse(rootDir, self._dirTree.currentCursor())
		self._recurse(rootDir, delegate, self._dirTree.currentCursor())



'''Traverses a JSON tree,
performing the given operations as requested.
'''
import os
from ...logging import Logger

class JsonRunner:
	'''Traverses a JSON tree,
	performing the given operations as requested.
	'''
	def __init__(self, rootDir, log, commandRunner, tree):
		assert rootDir
		self._rootDir = rootDir
		self._log = log
		assert isinstance(self._log, Logger)
		self._command = commandRunner
		assert tree
		self._tree = tree
		self._lastRunErrors = []

	def _addError(self, errorString):
		self._addError(errorString)
		self._lastRunErrors.append(errorString)

	def _dirChainToPath(self, dirChain):
		separator = os.path.sep
		return self._rootDir + separator + separator.join(dirChain)

	def _recurse(self, nodeName, nodeValue, dirChain, validateOnly):
		self._log.debug("Recursing on node {0}".format(nodeName))
		assert isinstance(nodeValue, dict), "Node {0} isn't a dictionary".format(nodeName)
		kIsRepoKey = "isRepo"
		#Recurse on parsed JSON:
		#Entry has a "isRepo" key && isRepo == True?
		isRepo = nodeValue[kIsRepoKey] if (kIsRepoKey in nodeValue) else False
		if isRepo:
			kRemotesKey = "remotes"
			#	Entry has a "remotes" key?
			if not kRemotesKey in nodeValue:
				#If not, mark error and return
				self._addError("Node {0} doesn't have a 'remote' entry".format(nodeName))
				return
			else:
				#Else, "remotes" has a "origin" key?
				remotes = nodeValue[kRemotesKey]
				kOriginKey = "origin"
				if not kOriginKey in remotes:
					#If not, mark error and return
					self._addError("Node {0} doesn't have an origin remote".format(nodeName))
					return
				else:
					#Else, git clone using "origin"'s string value
					repoAddress = remotes[kOriginKey]
					if not isinstance(repoAddress, str):
						self._addError("Key '{0}' of node '{1}' \
						isn't a Git repository \
						string".format(kOriginKey, nodeName))
						return
					if not validateOnly:
						repoDestinationPath = self._dirChainToPath(dirChain) + os.path.sep + nodeName
						#Skip if this directory already exists
						repoDestinationPathWithGit = repoDestinationPath + os.path.sep + ".git"
						if os.path.exists(repoDestinationPathWithGit):
							#TODO: should really check that expected remotes match
							self._log.warn("Node {0} appears to already be cloned, skipping".format(nodeName))
							return

						try:
							self._command.clone(repoAddress, repoDestinationPath)
						except RuntimeError as e:
							#If clone fails mark error and return
							self._addError("Couldn't clone repository \
							'{0}' to '{1}': {2}".format(repoAddress,
							repoDestinationPath,
							str(e)))
							return
					#Else, for each other key K named N in "remotes":
					for remoteName in [x for x in remotes if x != kOriginKey]:
						#Add K as remote with name N
						remoteAddress = remotes[remoteName]
						if not isinstance(remoteAddress, str):
							self._addError("Key '{0}' of node '{1}' \
							isn't a Git remote URL".format(remoteName, "{0}.{1}".format(nodeName, kRemotesKey)))
							return
						if not validateOnly:
							try:
								self._command.addRemote(repoDestinationPath, remoteName, remoteAddress)
							except RuntimeError as e:
								#If remote add fails mark error and return
								self._addError("Couldn't add remote \
								'{0}' at '{1}' to repository '{2}': {3}".format(remoteName,
								remoteAddress,
								repoDestinationPath,
								str(e)))
								return
		else:
			#Else, create subdirectory named after key
			subDirChain = list(dirChain)
			subDirChain.append(nodeName)
			if not validateOnly:
				try:
					self._command.mkdir(self._dirChainToPath(subDirChain))
				#If mkdir fails, mark error and return
				except IOError:
					self._addError("Couldn't make directory {0}, skipping node")
					return
			#For each node in this node's value:
			for subNodeName, subNodeValue in nodeValue.items():
				#Recurse with that node
				self._recurse(subNodeName, subNodeValue, subDirChain, validateOnly)

	def _recurseStart(self, validateOnly):
		#Is a "repos" key at the root?
		kReposKey = "repos"
		if not kReposKey in self._tree:
		#If not, error and halt traversal
			self._addError("File has no root '{0}' key, can't continue".format(kReposKey))
			return
		else:
			#	Else, is it a dictionary?
			repos = self._tree[kReposKey]
			if not isinstance(repos, dict):
				#If not, error and halt traversal
				self._addError("File has a '{0}' key, but it's not a dictionary".format(kReposKey))
				return
			else:
				#Else, is it empty?
				if not repos:
					#If yes, error and halt traversal
					self._addError("File has a '{0}' key, but it's empty".format(kReposKey))
					return
				else:
					self._log.debug("Found {0} entries, beginning recurse".format(len(repos)))
					#Else, for each pair:
					for nodeName, nodeValue in repos.items():
						#self._recurse with that pair
						self._recurse(nodeName, nodeValue, [], validateOnly)

	def run(self, validateOnly=False):
		'''Performs the traversal, cloning Git
		repositiories listed in the
		given parsed repo dictionary if
		validateOnly is False.
		Returns:
			* A list of errors in the tree,
			if any occurred.
		'''
		self._lastRunErrors.clear()
		#If the parsed dict doesn't exist or is empty,
		#that's an error!
		if not self._tree:
			self._addError("There doesn't appear to be \
			any parsed JSON data, can't perform clone")
		else:
			self._recurseStart(validateOnly)

		return self._lastRunErrors



'''Defines logger manager class.
'''
from .constants import LogLevel
from .log_entry import LogEntry

class LogFormatStrings(object):
	'''Defines format strings used
	by Logger._renderedLineForEntry().
	'''
	def __init__(self, fullFormatString,
	timecodeAndMessageFormatString,
	severityAndMessageFormatString,
	messageFormatString):
		self.fullFormatString = fullFormatString
		self.timecodeAndMessageFormatString = timecodeAndMessageFormatString
		self.severityAndMessageFormatString = severityAndMessageFormatString
		self.messageFormatString = messageFormatString

class RenderMode(object):
	'''Defines rendering settings
	used by Logger._renderedLineForEntry().
	'''
	def __init__(self, showTimecode=True, showLevelText=True, showLevelColor=True):
		'''Initializer for RenderModes.
		By default, timecodes are displayed, as are severity
		levels via both text and code.
		'''

		#If True, timecodes in the format hh:mm:ss:mmmm
		#are added to the rendered entry.
		self.showTimecode = showTimecode
		#If True, text-based severity codes
		#are added to the rendered entry.
		self.showLevelText = showLevelText
		#If True, color highlighting
		#is added to the rendered entry
		#to indicate severity.
		#Currently does nothing.
		self.showLevelColor = showLevelColor

	def formatStringForContext(self, logFormatStringsObject):
		'''Given a LogFormatStrings object,
		returns the format string that this
		render mode indicates should be used.
		'''
		assert isinstance(logFormatStringsObject, LogFormatStrings)
		if self.showTimecode:
			if self.showLevelText:
				return logFormatStringsObject.fullFormatString
			return logFormatStringsObject.timecodeAndMessageFormatString
		elif self.showLevelText:
			return logFormatStringsObject.severityAndMessageFormatString
		else:
			return logFormatStringsObject.messageFormatString


class Logger(object):
	'''Handles event logging and rendering for the application.
	'''
	#Format:
	#	0: time
	#	1: severity prefix
	#	2: message string
	_kFullFormatString = "[{0}] {1}{2}"
	_kTimecodeAndMessageFormatString = "[{0}] {2}"
	_kSeverityAndMessageFormatString = "{1}{2}"
	_kMessageFormatString = "{2}"
	_kLogFormatStrings = LogFormatStrings(_kFullFormatString,
	_kTimecodeAndMessageFormatString,
	_kSeverityAndMessageFormatString,
	_kMessageFormatString)

	def __init__(self, minimumPrintableLevel=LogLevel.Info,
	stdOutRenderContext=None, fileRenderContext=None):
		#Entries with a loglevel below this value
		#are not printed.
		self._minimumPrintableLevel = minimumPrintableLevel
		defaultLogFormatString = Logger._kLogFormatStrings.fullFormatString
		self._stdOutRenderFormatString = defaultLogFormatString \
		if stdOutRenderContext is None \
		else stdOutRenderContext.formatStringForContext(Logger._kLogFormatStrings)
		self._fileRenderFormatString = defaultLogFormatString \
		if fileRenderContext is None \
		else fileRenderContext.formatStringForContext(Logger._kLogFormatStrings)
		#The LogEntries stored by this logger.
		#When dumpEntriesToFile() is called,
		#all values in this list are saved to the
		#given log file.
		self._entries = []

	def _renderedLineForEntry(self, logEntry, formatString):
		'''Returns given log entry as a printable string.
		'''
		prefacesDict = {
			LogLevel.Debug : "(debug) ",
			LogLevel.Info : "",
			LogLevel.Warn : "(warning) ",
			LogLevel.Error : "(error!) "
		}
		#Color might not be available, but give it a try.
		logLevel = logEntry[2]
		logTime = logEntry[1]
		logMessage = logEntry[3]
		return formatString.format(logTime, prefacesDict[logLevel], logMessage)

	def _printEntry(self, logEntry):
		lineForEntry = self._renderedLineForEntry(logEntry, self._stdOutRenderFormatString)
		print(lineForEntry)

	def _addEntry(self, logLevel, message, extraData=None):
		logEntry = LogEntry.entry(logLevel, message, extraData=extraData)
		self._entries.append(logEntry)

		#Print this message to screen if
		#it's above minimum printable level.
		if logLevel >= self._minimumPrintableLevel:
			self._printEntry(logEntry)

	def dumpEntriesToFile(self, filePath):
		'''Saves all entries in the log
		to the specified file, appending to
		the file if it already exists.
		'''
		#Just appending, no need to read
		outFile = open(filePath, 'a')
		outFile.writelines("{0}\n".format(self._renderedLineForEntry(x, \
		self._fileRenderFormatString)) for x in self._entries)

	def entriesOfLevel(self, logLevel):
		'''Gets all log entries exactl matching
		the specified severity level.
		'''
		return [x for x in self._entries if x[2] == logLevel]

	def printLogEntries(self, entryList):
		'''Prints all log entries given.
		These have the same extraData as their original
		entries, but are new entries and go in the result log.
		'''
		for entry in entryList:
			self.logMessage(entry[2], entry[3], entry[5])

	def logMessage(self, logLevel, message, extraData=None):
		'''Logs a message with the given severity to the logger,
		printing it if LogLevel is at `logLevel` or below.
		'''
		assert logLevel >= 0 and logLevel < LogLevel.Count, \
		"Log level should be between 0 and {0}, is actually {1}".format(LogLevel.Count, logLevel)
		self._addEntry(logLevel, message, extraData)

	def debug(self, message, extraData=None):
		'''Logs a debug message to the logger,
		printing it if LogLevel is at Debug or below.
		'''
		self.logMessage(LogLevel.Debug, message, extraData)

	def info(self, message, extraData=None):
		'''Logs an info message to the logger,
		printing it if LogLevel is at Info or below.
		'''
		self.logMessage(LogLevel.Info, message, extraData)

	def warn(self, message, extraData=None):
		'''Logs a warning message to the logger,
		printing it if LogLevel is at Warning or below.
		'''
		self.logMessage(LogLevel.Warn, message, extraData)

	def error(self, message, extraData=None):
		'''Logs an error message to the logger,
		printing it if LogLevel is at Error or below.
		'''
		self.logMessage(LogLevel.Error, message, extraData)



'''Defines the main business logic of the application.
'''
import datetime
import json
import os
from ..logging import Logger, LogLevel, RenderMode
from .command import CommandRunner
from .json_traversal import JsonRunner
from .directory_traversal import DirectoryRunner, GitPullDelegate, BuildDictionaryDelegate, ListBranchDelegate

### Utilities
def pluralizeString(singularString, pluralString, number):
	'''Returns the singular string if number == 1,
	and the plural string otherwise.
	'''
	if number == 1:
		return singularString
	return pluralString

def absolutePath(pathString):
	'''Returns the absolute path of the given
	path string, if possible.
	'''
	return os.path.realpath(os.path.expanduser(pathString))

def fileSystemFriendlyDate():
	'''Returns a YYYY-MM-DD date
	that is valid for the common filesystems in use.
	'''
	return datetime.datetime.now().strftime("%Y-%m-%d")

class Program(object):
	'''Performs the business logic of the application.
	'''
	def __init__(self, args):
		self._args = args
		self._minLogLevel = LogLevel.Debug if args.debug else LogLevel.Info
		self.dryRun = args.dryrun
		showTimeCode = args.verbose
		stdOutRenderMode = RenderMode(showTimeCode)
		self._log = Logger(self._minLogLevel, stdOutRenderContext=stdOutRenderMode)
		self._parsedRepoDict = None
		self._command = CommandRunner(self._log, self.dryRun)

	def _logErrorAndRaise(self, message):
		self._log.error(message)
		raise RuntimeError(message)

	def _parseRepoFile(self, repoFilePath):
		'''Parses the given JSON repo file.
		Returns the parsed JSON as a dictionary.
		Raises:
			* Undecided error if the JSON is unparseable.
		'''
		jsonFile = None
		try:
			self._log.debug("Opening file '{0}'".format(repoFilePath))
			jsonFile = open(repoFilePath)
		except IOError:
			raise RuntimeError("Parsing file '{0}' failed".format(repoFilePath))

		assert jsonFile, \
		"File '{0}' reported successful load but doesn't actually exist".format(repoFilePath)
		return json.load(jsonFile)

	def _reportStatus(self, statusFormatString):
		'''Prints the results of the program run given
		a list of errors/warnings that occurred.
		'''
		errors = self._log.entriesOfLevel(LogLevel.Error)
		warnings = self._log.entriesOfLevel(LogLevel.Warn)
		numErrors = len(errors)
		numWarnings = len(warnings)
		resultStatus = LogLevel.Error
		statusString = ""
		#If there are any errors or warnings:
		if numErrors + numWarnings > 0:
			errorNoun = pluralizeString("error", "errors", numErrors)
			warningNoun = pluralizeString("warning", "warnings", numWarnings)
			#List those errors/warnings to terminal.
			if numErrors > 0:
				resultStatus = LogLevel.Error
				#Handle there being errors and possibly warnings...
				statusString = "failed with {0} {1}".format(numErrors, errorNoun)
				if numWarnings > 0:
					#Append warnings if they do exist
					statusString = statusString + " and {0} {1}".format(numWarnings, warningNoun)
				statusString = statusString + "!"
			elif numWarnings > 0:
				resultStatus = LogLevel.Warn
				#And there only being warnings.
				statusString = "complete, but {0} {1} were encountered!".format(numWarnings, warningNoun)
			else:
				raise RuntimeError("numErrors + numWarnings > 0, but numErrors \
				and numWarnings are both <= 0 somehow")
		else:
			#Else, print success message.
			resultStatus = LogLevel.Info
			statusString = "complete."

		self._log.logMessage(resultStatus, statusFormatString.format(statusString))

		#Print any errors or warnings...
		if numErrors + numWarnings > 0:
			self._log.info("")
			if numErrors > 0:
				self._log.info("Errors:")
				self._log.printLogEntries(errors)
			if numWarnings > 0:
				self._log.info("Warnings:")
				self._log.printLogEntries(warnings)

		#Save verbose log:
		#	Chronological list of events
		#	? Listing of errors in chronological order
		#	? Listing of warnings in chronological order
		#	Count of errors/warnings
		defaultLogPath = "gitup_log_{0}.log".format(fileSystemFriendlyDate())
		try:
			self._log.dumpEntriesToFile(defaultLogPath)
		except IOError:
			print("Couldn't open log file '{0}', this run wasn't logged!".format(defaultLogPath))

	def _jsonRun(self, statusFormatString, validateOnly, runFailedString, runRaisedExceptionString):
		'''Given a JSON file, creates all listed subdirectories
		and clones the given origin repositories, adding any
		remotes listed as well.
		'''
		repoFilePath = self._args.target
		outPath = self._args.out
		#Parse the JSON.
		try:
			self._parsedRepoDict = self._parseRepoFile(repoFilePath)
			#Actually do the cloning...
			cloneTreeRunner = JsonRunner(outPath, self._log, self._command, self._parsedRepoDict)
			treeErrors = cloneTreeRunner.run(validateOnly)
			if treeErrors:
				self._log.error(runFailedString)
		except RuntimeError as e:
			self._log.error(runRaisedExceptionString + ": {0}".format(str(e)), e)

		self._reportStatus(statusFormatString)

	def _cloneRepos(self):
		'''Clones repos and adds remotes
		listed in the JSON file.
		'''
		self._jsonRun("Repo cloning {0}", False, \
		"Cloning was unsuccessful",
		"Fatal error while performing clone")

	def _validateFile(self):
		'''Validates that the JSON file has no
		errors.
		'''
		#Args will contain flags we need.
		repoFilePath = self._args.target
		self._jsonRun("Repo validation {0}", True, \
		"File '{0}' failed validation".format(repoFilePath), \
		"Fatal error while validating file '{0}'".format(repoFilePath))

	def _pullRepos(self):
		treeRootPath = self._args.target
		try:
			directoryRunner = DirectoryRunner(self._log)
			directoryRunner.run(treeRootPath, GitPullDelegate(self._command))
		except RuntimeError as e:
			self._log.error("Fatal error while pulling repositories: {0}".format(str(e)), e)

		self._reportStatus("Recursive pull on {0} {1}".format(treeRootPath, "{0}"))

	def _exportRepos(self):
		treeRootPath = self._args.target
		pathToFix = self._args.out
		#Use a default path if one was given
		if not pathToFix:
			defaultFileName = "gitup_export_{0}.json".format(fileSystemFriendlyDate())
			pathToFix = ".{0}{1}".format(os.path.sep, defaultFileName)
		destinationFilePath = absolutePath(pathToFix)
		canRunExport = True
		#Check that the destination doesn't already exist;
		#we don't overwrite unless -f is passed
		if not self.dryRun:
			if os.path.isdir(destinationFilePath):
				self._log.error(("Destination file '{0}' is a "
				"directory, nice try").format(destinationFilePath))
				canRunExport = False
			else:
				if os.path.exists(destinationFilePath):
					if not self._args.force:
						self._log.error(("Destination file '{0}' "
						"already exists, halting (use -f to "
						"overwrite anyway)").format(destinationFilePath))
						canRunExport = False
		if canRunExport:
			try:
				dictionaryDelegate = BuildDictionaryDelegate(self._log)
				directoryRunner = DirectoryRunner(self._log)
				directoryRunner.run(treeRootPath, dictionaryDelegate)
				if not self.dryRun:
					self._log.debug("Exporting to file '{0}'".format(destinationFilePath))
					destinationFile = open(destinationFilePath, "w")
					#Dump the results to the destination file.
					json.dump(directoryRunner.directoryStructure(), destinationFile, indent=4)
				else:
					self._log.debug("Dry run: export to file '{0}'".format(destinationFilePath))
					self._log.debug(("Dry run: data to export: "
					"'{0}'").format(directoryRunner.directoryStructure()))
			except RuntimeError as e:
				self._log.error("Fatal error while exporting directory tree: {0}".format(str(e)), e)

		self._reportStatus(("Export of directory tree starting "
		"at '{0}' to '{1}' {2}").format(treeRootPath, destinationFilePath, "{0}"))

	def _getBranches(self):
		treeRootPath = self._args.target
		try:
			directoryRunner = DirectoryRunner(self._log)
			directoryRunner.run(treeRootPath, ListBranchDelegate(self._log, self._command))
		except RuntimeError as e:
			self._log.error("Fatal error while getting repository branches: {0}".format(str(e)), e)

		self._reportStatus("Get branches of {0} {1}".format(treeRootPath, "{0}"))

	def run(self):
		#Check subparser to see what should be done.
		command = self._args.command
		if command == "clone":
			self._cloneRepos()
		elif command == "validate":
			self._validateFile()
		elif command == "pull":
			self._pullRepos()
		elif command == "export":
			self._exportRepos()
		elif command == "getbranches":
			self._getBranches()
		else:
			raise RuntimeError("Invalid command {0}".format(command))



class DirectoryTree(object):
	'''Builds a dictionary representing
	the repo tree that can then be
	converted into JSON.
	'''
	class Cursor(object):
		def __init__(self, name, location):
			self.name = name
			self.location = location

		def cursorForSubNode(self, nodeName):
			if not nodeName in self.location:
				raise RuntimeError(("Cursor '{0}' doesn't have child "
				"node '{1}', can't enter that node").format(self.name, nodeName))
			newLocation = self.location[nodeName]
			if not isinstance(newLocation, dict):
				raise RuntimeError(("Child node '{0}' on cursor '{1}' "
				"isn't a dictionary, can't place a cursor on "
				"that node").format(nodeName, self.name))
			return DirectoryTree.Cursor(nodeName, self.location[nodeName])

		def addValue(self, key, value):
			'''Adds the given value to the given
			key below the current cursor.
			Raises RuntimeError if the key already
			exists.
			'''
			if key in self.location:
				raise RuntimeError(("Cursor '{0}' already has child "
				"node '{1}' mapped to '{2}'").format(self.name, key, self.location[key]))
			self.location[key] = value

		def addSubNode(self, subNodeName):
			'''Adds an empty node below the current cursor,
			but doesn't actually enter it.
			Raises RuntimeError if the node already
			exists.
			'''
			self.addValue(subNodeName, {})

	def __init__(self, rootNodeName):
		self._rootNodeName = rootNodeName
		assert rootNodeName, "Root node name can't be empty"
		self.directoryStructure = {self._rootNodeName: {}}
		self._cursor = DirectoryTree.Cursor(self._rootNodeName, self.directoryStructure[self._rootNodeName])

	def resetState(self):
		self.directoryStructure = {self._rootNodeName: {}}
		self._cursor = DirectoryTree.Cursor(self._rootNodeName, self.directoryStructure[self._rootNodeName])

	def currentCursor(self):
		return self._cursor



'''Contains functions that
map to command-line operations.
'''
import os
import subprocess
from ..logging import Logger

class CommandRunner(object):
	'''Contains functions that
	map to command-line operations.
	'''
	def __init__(self, log, dryRun):
		self._log = log
		self._dryRun = dryRun
		assert isinstance(self._log, Logger)

	def _enterDirAndExecute(self, commandLabel, runDirectory, command):
		'''Changes directory to `runDirectory`,
		runs `command`, then changes directory back to the cwd.
		Returns: Any standard output from the command execution.
		Parameters:
		  * `commandLabel`: a human-readable name for the command
		  * `runDirectory`: the directory to run the command in
		  * `command`: the actual command as a list of strings.
		'''
		originalCwd = os.getcwd()
		os.chdir(runDirectory)
		resultCode = 0
		result = None
		try:
			result = subprocess.check_output(command)
		except subprocess.CalledProcessError as e:
			resultCode = e.returncode
		os.chdir(originalCwd)
		if resultCode:
			raise RuntimeError("{0} failed with code {1}".format(commandLabel, resultCode))
		return result

	def clone(self, repoAddress, destination):
		'''Clones the given address if
		self._dryRun is not enabled.
		'''
		if self._dryRun:
			self._log.debug("Dry run: clone {0} to {1}".format(repoAddress, destination))
			return
		else:
			self._log.debug("Cloning repo {0} to {1}".format(repoAddress, destination))
			result = subprocess.call(["git", "clone", repoAddress, destination])
			if result:
				raise RuntimeError("clone failed with code {0}".format(result))

	def addRemote(self, repoPath, remoteName, remoteUrl):
		if self._dryRun:
			self._log.debug(("Dry run: add remote "
			"{0} at {1} to repo {2}").format(remoteName, remoteUrl, repoPath))
			return
		else:
			self._log.debug(("Adding remote {0} at {1} "
			"to repo {2}").format(remoteName, remoteUrl, repoPath))
			self._enterDirAndExecute("addRemote", repoPath, ["git", "remote", "add", remoteName, remoteUrl])

	def pull(self, repoPath):
		'''Performs `git pull origin`
		on the given repo path.
		'''
		if self._dryRun:
			self._log.debug("Dry run: git pull origin for {0}".format(repoPath))
			return
		else:
			self._log.debug("Pulling origin for repo {0}".format(repoPath))
			self._enterDirAndExecute("pull", repoPath, ["git", "pull", "origin"])

	def getCurrentBranch(self, repoPath):
		'''Returns the current branch for the given
		repo path, if it exists.
		Raises RuntimeError if repoPath isn't
		actually a Git repository or can't be accessed.
		'''
		kCommand = ["git", "symbolic-ref", "--short", "HEAD"]
		if self._dryRun:
			self._log.debug("Dry run: git pull origin for {0}".format(repoPath))
			return
		else:
			self._log.debug("Getting current branch for repo {0}".format(repoPath))
			branchOutput = self._enterDirAndExecute("get-branch", repoPath, kCommand).decode()
			return branchOutput.strip('\n')

	def mkdir(self, dirAbsPath):
		'''Makes the given directory if
		self._dryRun is not enabled.
		'''
		if self._dryRun:
			self._log.debug("Dry run: mkdir {0}".format(dirAbsPath))
			return
		else:
			self._log.debug("Making directory {0}".format(dirAbsPath))
			os.makedirs(dirAbsPath, exist_ok=True)



'''Defines log entry format for the
Logger class.
'''
import datetime
from .constants import LogLevel, LogEntryType

class LogEntry:
	'''LogEntries are just tuples of the following:
	0: entry ID
	1: time entry was made
	2: log severity level
	3: the actual message, as a string
	4: message type; use this to determine what the data pointer contains
	5: an extra data pointer; typically this is null
	'''
	_nextEntryId = 0

	@classmethod
	def entry(cls, logLevel, messageString, messageType=LogEntryType.GeneralText, extraData=None):
		'''Generates a LogEntry tuple from the given parameters.
		'''
		assert logLevel >= 0 and logLevel < LogLevel.Count, \
		"Log level should be between 0 and {0}, is actually {1}".format(LogLevel.Count, logLevel)
		assert messageType >= 0 and messageType < LogEntryType.Count, \
		"messageType should be between 0 and {0}, is actually {1}".format(LogEntryType.Count, messageType)

		entryTime = datetime.datetime.now()
		logEntry = (cls._nextEntryId, entryTime, logLevel, messageString, messageType, extraData)
		cls._nextEntryId += 1
		return logEntry



'''Setup file for gittools package.
'''
from setuptools import setup, find_packages

setup(
	# Application name:
	name="jtitor_gittools",
	# Version number (initial):
	version="0.1.2",
	# Application author details:
	author="jTitor",
	author_email="name@addr.ess",
	# Packages
	packages=find_packages(),
	# Include additional files into the package
	include_package_data=True,
	# Details
	url="http://pypi.python.org/pypi/jtitor_gittools/0.1.2/",
	license="MIT",
	description="Batches common bulk operations on git repositories, but privately listed since it's not quite done.",
	# long_description=open("README.txt").read(),
	# Dependent packages (distributions)
	#install_requires=[
	#	"flask",
	#],
	entry_points={
		'console_scripts': ['gittools=app.gittools:main']
	}
)



from os import path
import re

from ..command import CommandRunner
from ...logging import Logger

class GitPullDelegate(object):
	'''Performs `git pull origin`
	on all repos found; does not recurse
	on any repos found.
	'''
	def onStartTraverse(self, treeRootPath, cursor):
		pass

	def onRepoFound(self, treeRootPath, cursor):
		#Try to pull origin repo
		self._command.pull(treeRootPath)

	def onSubdirectoryFound(self, subDirectoryPath, cursor):
		#no-op, can't do anything with a non-repo
		pass

	def onEnteredNode(self, nodePath, cursor):
		#no-op
		pass

	def __init__(self, commandRunner):
		self._command = commandRunner
		assert isinstance(self._command, CommandRunner)

class ListBranchDelegate(object):
	'''Prints the currently active branch of any
	repos found; does not recurse
	on any repos found.
	'''
	def __init__(self, log, commandRunner):
		self._log = log
		assert isinstance(self._log, Logger)
		self._command = commandRunner
		assert isinstance(self._command, CommandRunner)

	def onStartTraverse(self, treeRootPath, cursor):
		pass

	def onRepoFound(self, treeRootPath, cursor):
		#Try to pull origin repo
		currentBranch = self._command.getCurrentBranch(treeRootPath)
		self._log.info("Repo '{0}' is on branch '{1}'".format(treeRootPath, currentBranch))

	def onSubdirectoryFound(self, subDirectoryPath, cursor):
		#no-op, can't do anything with a non-repo
		pass

	def onEnteredNode(self, nodePath, cursor):
		#no-op
		pass

kReposKey = "repos"
class BuildDictionaryDelegate(object):
	'''Builds a dictionary representing
	the repo tree that can then be
	converted into JSON.
	'''

	def __init__(self, log):
		self._log = log
		assert isinstance(self._log, Logger)

	def _getRepos(self, repoPath):
		#Open the repo's /.git/config
		expectedConfigPath = repoPath + "/.git/config"
		results = {}
		try:
			with open(expectedConfigPath, "r") as configFile:
				configText = configFile.read()
				#Pull all of the repo files as a regex.
				#Regex strings should be raw strings?
				kRemoteRegexPattern = r'remote "(.*)".*\n\s*url = (.*)'
				#Remember to use multiline mode too.
				matches = re.compile(kRemoteRegexPattern, re.M).findall(configText)
				if not matches:
					self._log.error(("No matches found on repo "
					"config file {0}, regex pattern might be wrong").format(expectedConfigPath))
				else:
					#Now read each of the matches and turn those
					#into dictionary entries...
					self._log.debug(("Parsing repo config file {0}").format(expectedConfigPath))
					for m in matches:
						remoteName = m[0]
						remoteUrl = m[1]
						if not remoteName or not remoteUrl:
							self._log.error(("One of the matches "
							"in this config file is invalid. "
							"Key: '{0}', URL: '{1}'").format(remoteName, remoteUrl))
						else:
							#Make sure this repo doesn't
							#already have a remote...
							if remoteName in results:
								self._log.error(("Repo already "
								"has a value for remote key '{0}'").format(remoteName))
							else:
								#Now add that key!
								results[remoteName] = remoteUrl
		except IOError as e:
			self._log.error(("Failed to open repo config file at "
			"'{0}': {1}").format(expectedConfigPath, str(e)))
		return results

	def onStartTraverse(self, treeRootPath, cursor):
		pass

	def onRepoFound(self, treeRootPath, cursor):
		kIsRepoKey = "isRepo"
		kRemotesKey = "remotes"
		#Add a "isRepo" key
		cursor.addValue(kIsRepoKey, True)
		cursor.addSubNode(kRemotesKey)
		#Add the root of the "remotes" key
		remotesCursor = cursor.cursorForSubNode(kRemotesKey)
		#Get all of the remotes
		remotesFound = self._getRepos(treeRootPath)
		for key, value in remotesFound.items():
			#Add each to the remotes key
			remotesCursor.addValue(key, value)

	def onSubdirectoryFound(self, subDirectoryPath, cursor):
		pass

	def onEnteredNode(self, nodePath, cursor):
		pass



'''Given a JSON file, attempts to clone the given repositories.
Requires Python 3!
'''
import argparse
from .program import Program

def main():
	'''Entry point for program.
	'''

	parser = argparse.ArgumentParser(prog="gittools",
	description="Batches common Git operations.")
	parser.add_argument("-d", "--debug",
						action="store_true",
						help=("Show debug entries in standard output."
						"Result logs always include debug info."))
	parser.add_argument("--dryrun",
						action="store_true",
						help=("List actions that would be performed, but"
						"do not actually perform them if they would"
						"affect the filesystem."))
	parser.add_argument("-v", "--verbose",
						action="store_true",
						help=("Include timecodes in standard output"
						"and show verbose output that would normally be hidden."))

	#Add subparsers:
	subparsers = parser.add_subparsers(dest="command")
	#clone command
	cloneParser = subparsers.add_parser("clone",
	help=("Reconstructs the directory tree described in a JSON"
	"file, cloning all Git repositories listed and adding any"
	"listed remotes."))
	cloneParser.add_argument("target",
	help="Path to a JSON file describing the repositories to clone.")
	cloneParser.add_argument("-o", "--out", default=".",
	help="Path to the output directory.")
	#validate command
	validateParser = subparsers.add_parser("validate",
	help=("Checks that the given JSON"
	"file is valid for use in `gittools clone`."))
	validateParser.add_argument("target",
	help="Path to a JSON file describing the repositories to clone.")
	#TODO: Can this be removed?
	validateParser.set_defaults(out=".")
	#pull command
	pullParser = subparsers.add_parser("pull",
	help=("Performs `git pull origin` on all Git repositories"
	"recursively below the given directory. Does not"
	"recurse on repository directories."))
	pullParser.add_argument("target",
	help="Path to the top of the directory tree to update.")
	#export command
	exportParser = subparsers.add_parser("export",
	help=("Exports all Git repositories"
	"recursively below the given directory to a JSON dictionary."
	"Does not recurse on repository directories."))
	exportParser.add_argument("target",
	help="Path to the top of the directory tree to export.")
	exportParser.add_argument("-o", "--out",
	help=("Path to the output dictionary file. If omitted, writes"
	"to the current working directory."))
	exportParser.add_argument("-f", "--force",
	action="store_true",
	help=("Overwrites the output dictionary file if it already "
	"exists."))
	#get-branches command
	getBranchesParser = subparsers.add_parser("getbranches",
	help=("Lists the current branch of all Git repositories"
	"recursively below the given directory."
	"Does not recurse on repository directories."))
	getBranchesParser.add_argument("target",
	help="Path to the top of the directory tree to list.")

	args = parser.parse_args()
	if not args.command:
		args = parser.parse_args(["--help"])
	Program(args).run()

if __name__ == "__main__":
	main()
